/*
 * Copyright (c) 1995, 1996, 1997 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include <string.h>
#include <assert.h>
#include <mom/compiler.h>
#include <mom/libmint.h>
#include <mom/c/libcast.h>
#include <mom/c/libpres_c.h>
#include <mom/c/pbe.hh>
#include "data_types.hh"
#include "hash_const.hh"

/* This is a hash function of sorts.
   It takes a list of possibilities (the values to hash),
   and creates code that will assign var
   the index of the value encountered. 
   Example:
   "operation1", "operation2", or "operation3"
   are the constants we wish to switch on.
   Since you can't switch on a string,
   we have to build a variable to switch on.
   
   switch(str[0]) {
   case 'o':
       switch(str[1]) {
       case 'p':
       ...
           switch(str[9]) {
	   case '1':
	      var = 0;
	      break;
	   case '2':
	      var = 1;
	      break;
	   case '3':
	      var = 2;
	      break;
	   default:
	      var = -1;
	   }
       default:
           var = -1;
       }
   default:
       var = -1;
   }
   
   Now we can switch on var...
   This code should build the 'hash' function
   */

/*****************************************************************************/

/* Previously, the `hash' method was designed only to hash complex values into
   a range of integer "cookie" values.  The code determined the cookie values
   for itself.
   
   But now, the code has been modified to use the "functor" mechanism.  Each
   value in the input domain is associated with a functor, which is expected to
   build the C code that will be executed when the corresponding value is
   recognized.
   
   The `hash_const_cookie_functor' is a functor that replicates the old
   "cookie" behavior.  When one of these functors is executed, it outputs an
   assignment statement that stores the cookie value into the cookie variable.
   
   Soon, however, this class (and the code that uses it) is expected to become
   obsolete.  The reason for cookifying our values is so that subsequent code
   can switch on the cookies.  But now, that subsequent code can be inlined
   directly into the `switch' generated by `hash()'.  So there's no reason to
   generate cookies at all.
   */

struct hash_const_cookie_functor : public functor {
	hash_const_cookie_functor(cast_expr var, int val) :
		variable(var), value(val) {}
	
	virtual void func(mu_state *must);
	
	cast_expr variable;
	int value;
};

void hash_const_cookie_functor::func(mu_state *must)
{
	must->add_stmt(cast_new_stmt_expr(
		cast_new_expr_assign(variable,
				     cast_new_expr_lit_int(value,
							   CAST_MOD_SIGNED)
				     )
		));
}


/*****************************************************************************/

/* encodes 4 chars into a big-endian unsigned 32 bit int */
static unsigned int FourCharsTo32UInt(char c1, char c2, char c3, char c4)
{
  return (((((unsigned int)c1 << 8) + c2) << 8) + c3 << 8) + c4;
}

/* encodes 2 chars into a big-endian unsigned 16 bit int */
static unsigned short int TwoCharsToUShort(char c1, char c2)
{
  return ((unsigned short int)c1 << 8) + c2;
}

char flick_case_InttoCharResult[5];

/* decodes a big-endian unsigned 32 bit int into a string of 4 chars */
static char *UInt32ToFourChars(unsigned int x)
{
  for (int i = 0; i <= 3; i++) {
	  flick_case_InttoCharResult[3 - i] = x & 0xFF;
	  x = x >> 8;
  }
  flick_case_InttoCharResult[4] = 0;
  return flick_case_InttoCharResult;
}

/* decodes a big-endian unsigned 16 bit int into a string of 2 chars */
static char *UShortToTwoChars(unsigned short int x)
{
  flick_case_InttoCharResult[0] = x >> 8;
  flick_case_InttoCharResult[1] = x & 0xFF;
  flick_case_InttoCharResult[2] = 0;
  return flick_case_InttoCharResult;
}

void 
hash_const::hash()
{
  mu_memory_allocator_state *enter_case_state;
  cast_stmt last_block, default_block, temp;
  int *identical, groupcount = 0;
  unsigned int i, j;
  // the most char's we can lump together as an int
  unsigned int mostpackable = must->get_most_packable();
  mint_ref type_index = must->pres->mint.defs.defs_len;

  // This stuff is used for each individual group
  functor **suc_funs;
  int num_in_group, curpos, flag;
  data_type *dom, // The new domain (for this group of similar values)
    *optimal_dom; // An intermediate domain for grouping char_arrays into longs
  cast_expr case_val;

  // XXX This is NOT completely correct.  This should ONLY optimize for packed arrays of char!
 
  // Find either 2 or 4 consecutive chars for optimizing into 'int' or 'short'
  for (i = 0; i < (unsigned int)count; i++) {
    if (domain[i][0].type != data_field::TYPE_CHAR) {
      mostpackable = 0;
      break;
    }
    for (j = 1; (j < mostpackable) && (j < (unsigned int)domain[i].count()); 
	 j++)
      if (domain[i][j].type != data_field::TYPE_CHAR) 
	    break; // quit the for j loop
    if (j < mostpackable) 
      mostpackable = j;
  }


  if (mostpackable >= 2) {
    // Optimization will be used
    optimal_dom = new data_type[(unsigned int)count];
    if (mostpackable >= 4) {
      // found 4 packed chars. Optimize to be a 4 byte int.
      for (i = 0; i < (unsigned int)count; i++) {
        optimal_dom[i] = data_type(*new data_field(FourCharsTo32UInt(
					    domain[i][0].u.c,
					    domain[i][1].u.c,
					    domain[i][2].u.c,
					    domain[i][3].u.c)));
	assert(optimal_dom[i][0].category == data_field::TYPE_LITERAL 
	       || optimal_dom[i][0].category == data_field::TYPE_SYMBOLIC);
	for (j = 4; j < (unsigned int)domain[i].count(); j++)
	  optimal_dom[i] = optimal_dom[i] + domain[i][j];
      }
    } else {
      // found 2 packed chars. Optimize to be a 2 byte short.
      for (i = 0; i < (unsigned int)count; i++) {
	optimal_dom[i] = data_type(
 	  *new data_field(TwoCharsToUShort(domain[i][0].u.c,
					   domain[i][1].u.c)));
	assert(optimal_dom[i][0].category == data_field::TYPE_LITERAL 
	       || optimal_dom[i][0].category == data_field::TYPE_SYMBOLIC);
	for (j = 2; j < (unsigned int)domain[i].count(); j++)
	  optimal_dom[i] = optimal_dom[i] + domain[i][j];
      }
    }
  } else
    // found a single field.  No optimization possible.
    optimal_dom = domain;

  // unmarshal the variable to switch on
  switch (optimal_dom[0][0].type) {
  case data_field::TYPE_INT:
    switch_type = cast_new_prim_type(CAST_PRIM_INT, CAST_MOD_SIGNED);
    switch_var = must->add_temp_var("switch_tag", switch_type, CAST_SC_AUTO);
    type_index = must->pres->mint.standard_refs.signed32_ref;
    break;
    
  case data_field::TYPE_UINT:
    switch_type = cast_new_prim_type(CAST_PRIM_INT, CAST_MOD_UNSIGNED);
    switch_var = must->add_temp_var("switch_tag", switch_type, CAST_SC_AUTO);
    type_index = must->pres->mint.standard_refs.unsigned32_ref;
    break;
    
  case data_field::TYPE_USHORT:
    switch_type = cast_new_prim_type(CAST_PRIM_INT, CAST_MOD_UNSIGNED
				                    | CAST_MOD_SHORT);
    switch_var = must->add_temp_var("switch_tag", switch_type, CAST_SC_AUTO);
    type_index = must->pres->mint.standard_refs.unsigned16_ref;
    break;
    
  case data_field::TYPE_FLOAT:
    switch_type = cast_new_prim_type(CAST_PRIM_FLOAT, 0);
    switch_var = must->add_temp_var("switch_tag", switch_type, CAST_SC_AUTO);
    type_index = must->pres->mint.standard_refs.float32_ref;
    break;
    
  case data_field::TYPE_CHAR:
    switch_type = cast_new_prim_type(CAST_PRIM_CHAR, 0);
    switch_var = must->add_temp_var("switch_tag", switch_type, CAST_SC_AUTO);
    type_index = must->pres->mint.standard_refs.char8_ref;
    break;
    
  default:
    panic("Incorrect type to hash_const::hash()\n");
  }
  assert(type_index < (signed int) must->pres->mint.defs.defs_len);

  must->mu_mapping_direct(switch_var, switch_type, type_index);
  must->break_chunk();
  // At this point, we take a "snapshot" of our glob/chunk memory allocation
  // state.  Every `case' in the `switch' will start from this saved memory
  // state.
  enter_case_state = must->memory_allocator_state();
  // Start a new block
  last_block = must->c_block;
  must->c_block = 0;
  
  // find identical cases in the optimal_dom and group them
  identical = new int[count];
  for (i = 0; i < (unsigned int)count; i++) 
    identical[i] = -1;
  for (i = 0; i < (unsigned int)count; i++) {
    if (identical[i] < 0) {
      for (j = i+1; j < (unsigned int)count; j++) 
	if (identical[j] < 0 && (optimal_dom[i][0]==optimal_dom[j][0]))
	  identical[j] = groupcount;
      identical[i] = groupcount++;
    }
  }
	
  // Now step thru each unique case and build a case statement for it
  // If we're done with the identifier (length is 1) then set the value
  // Otherwise, make a new hash object to hash the similar values
  
  for (i = 0; i < (unsigned int)groupcount; i++) {
    num_in_group = 0;
    flag = 0;
    curpos = 0;
    // Find the number of members in this group
    for (j = 0; j < (unsigned int)count; j++)
      num_in_group += ((unsigned int)identical[j] == i);

    suc_funs = new functor *[num_in_group];
    dom = new data_type[num_in_group];
    // Build the list of possibilities for this branch
    for (j = 0; j < (unsigned int)count; j++) {
      if ((unsigned int)identical[j] == i) {
	suc_funs[curpos] = success_functors[j];
	// Eat the first identifier - it will have already been 'case'd
	dom[curpos++] = optimal_dom[j].crop();
	if (!flag) {
		switch(optimal_dom[j][0].category) {
		case data_field::TYPE_LITERAL:
			switch (optimal_dom[j][0].type) {
			case data_field::TYPE_INT:
				case_val = cast_new_expr_lit_int(
					optimal_dom[j][0].u.i,
					CAST_MOD_SIGNED);
				break;
			case data_field::TYPE_UINT:
				case_val = cast_new_expr_lit_int(
					(signed long long)optimal_dom[j][0].u.u,
					CAST_MOD_UNSIGNED);
				if (mostpackable >= 4)
					case_val
					  = cast_new_expr_call_1
					  (cast_new_expr_name
					   (flick_asprintf
					    ("/* %s */ flick_%s_make_big_endian_const_unsigned32",
					     UInt32ToFourChars(optimal_dom[j][0].u.u),
					     must->get_encode_name())),
					   case_val);
				break;
			case data_field::TYPE_USHORT:
				case_val = cast_new_expr_lit_int(
					optimal_dom[j][0].u.s,
					CAST_MOD_UNSIGNED | CAST_MOD_SHORT);
				if (mostpackable >= 2)
					case_val
					  = cast_new_expr_call_1
					  (cast_new_expr_name
					   (flick_asprintf
					    ("/* %s */ flick_%s_make_big_endian_const_unsigned16",
					     UShortToTwoChars(optimal_dom[j][0].u.s),
					     must->get_encode_name())),
					   case_val);
				break;
			case data_field::TYPE_CHAR:
				case_val = cast_new_expr_lit_char(
					optimal_dom[j][0].u.c,
					0);
				break;
			case data_field::TYPE_FLOAT:
				case_val = cast_new_expr_lit_float(
					optimal_dom[j][0].u.f);
				break;
			default:
				panic("BAD error in mu_hash_const.cc\n");
			}
			break;
			
		case data_field::TYPE_SYMBOLIC:
			case_val = cast_new_expr_name(optimal_dom[j][0].u.name);
			break;
			
		default:
			panic("BAD error in mu_hash_const.cc\n");
		}
		flag = 1;
	}
      }
    }
  
    must->set_memory_allocator_state(enter_case_state);
    add_case(dom, suc_funs, num_in_group, case_val);
    delete[] dom;
    delete[] suc_funs;
  }
  
  // Add the failure case (default)
  temp = must->c_block;
  must->c_block = 0;
  must->set_memory_allocator_state(enter_case_state);
  failure_functor->func(must);
  must->add_stmt(cast_new_break());
  default_block = must->c_block;
  must->c_block = temp;
  must->add_stmt(cast_new_default(default_block));
  delete enter_case_state;
  
  // Now actually add the switch statement to the function body
  temp = must->c_block;
  must->c_block = last_block;
#if 0
  if (mostpackable >= 4)
    must->add_stmt(cast_new_switch(cast_new_expr_call_1
				   (cast_new_expr_name("BIG_ENDIAN_32CONST"),
				    switch_var), temp));
  else if (mostpackable >= 2)
    must->add_stmt(cast_new_switch(cast_new_expr_call_1
				   (cast_new_expr_name("BIG_ENDIAN_16CONST"),
				    switch_var), temp));
  else 
#endif
    must->add_stmt(cast_new_switch(switch_var, temp));
}

void 
hash_const::add_case(data_type *d, functor **suc_funs, int count, cast_expr case_val) 
{
  // Add the case statement
  cast_stmt temp_block = must->c_block;
  
  // Add the stuff to do if this is the case within a {} of code
  must->c_block = 0;
  if (!count) {}
  else if(count == 1 && d->count() == 0) {
	  suc_funs[0]->func(must);
  } else {
    // We need to add the nested case for this set of values
    hash_const next_slot(d, suc_funs, failure_functor, count, this);
    next_slot.hash();
  }
  must->add_stmt(cast_new_break());
  
  // Get the constant value to select
  cast_stmt this_case = cast_new_case(case_val, must->c_block);
  must->c_block = temp_block;
  must->add_stmt(this_case);
}

hash_const::hash_const(data_type *d, functor **succ_funs, functor *fail_fun,
		       int count, hash_const *hc) :
	must(hc->must),
	domain(d), 
	count(count), 
	success_functors(succ_funs),
	failure_functor(fail_fun)
{
}

void 
mu_state::mu_hash_const(mint_const *domain,	// the 'domain' of inputs
			int domain_size,		// the # of inputs
			mint_ref const_type,	// the type of the inputs
			cast_expr var)		// the variable to hash to
{
  hash_const *hash_func;

  // must be a real reference and in range
  assert (const_type >= 0);
  assert (const_type < (signed int)pres->mint.defs.defs_len);
  
  switch (pres->mint.defs.defs_val[const_type].kind) {
  case MINT_INTEGER:
    hash_func = new hash_const_int(this, domain, domain_size, var);
    break;
		
  case MINT_CHAR: 
    hash_func = new hash_const_char(this, domain, domain_size, var);
    break;

  case MINT_FLOAT: 
    hash_func = new hash_const_float(this, domain, domain_size, var);
    break;

  case MINT_ARRAY:
    hash_func = new hash_const_array(this, domain, domain_size, var);
    break;
    
  case MINT_STRUCT:
    hash_func = new hash_const_struct(this, domain, domain_size, var);
    break;
    
  default:
    panic("mu_hash_const: unimplemented type for a constant...\n");
  }	
  hash_func->hash();
}
 
hash_const_int::hash_const_int(mu_state *m, mint_const *d, 
			       int domain_size, cast_expr var)
{
  must = m;
  count = domain_size;
  // build the data_type domain...
  domain = new data_type[count];
  success_functors = new functor *[count];
  for (int i = 0; i < count; i++) {
    if (d[i]->kind != MINT_CONST_INT)
      panic("Incorrect type in a domain of constants (int expected)...\n");
    success_functors[i] = new hash_const_cookie_functor(var, i);
    domain[i] = build(d[i]->mint_const_u_u.const_int);
  }
  failure_functor = new hash_const_cookie_functor(var, -1);
}

hash_const_char::hash_const_char(mu_state *m, mint_const *d,
				 int domain_size, cast_expr var)
{
  must = m;
  count = domain_size;
  // build the data_type domain...
  domain = new data_type[count];
  success_functors = new functor *[count];
  for (int i = 0; i < count; i++) {
    if (d[i]->kind != MINT_CONST_CHAR)
      panic("Incorrect type in a domain of constants (char expected)...\n");
    success_functors[i] = new hash_const_cookie_functor(var, i);
    domain[i] = build(d[i]->mint_const_u_u.const_char);
  }
  failure_functor = new hash_const_cookie_functor(var, -1);
}

hash_const_float::hash_const_float(mu_state *m, mint_const *d, 
				   int domain_size, cast_expr var)
{
  must = m;
  count = domain_size;
  // build the data_type domain...
  domain = new data_type[count];
  success_functors = new functor *[count];
  for (int i = 0; i < count; i++) {
    if (d[i]->kind != MINT_CONST_FLOAT)
      panic("Incorrect type in a domain of constants (float expected)...\n");
    success_functors[i] = new hash_const_cookie_functor(var, i);
    domain[i] = build(d[i]->mint_const_u_u.const_float);
  }
  failure_functor = new hash_const_cookie_functor(var, -1);
}

hash_const_array::hash_const_array(mu_state *m, mint_const *d, 
				   int domain_size, cast_expr var)
{
  must = m;
  count = domain_size;
  // build the data_type domain...
  domain = new data_type[count];
  success_functors = new functor *[count];
  for (int i = 0; i < count; i++) {
    if (d[i]->kind != MINT_CONST_ARRAY)
      panic("Incorrect type in a domain of constants...\n");
    success_functors[i] = new hash_const_cookie_functor(var, i);
    domain[i] = build(d[i]->mint_const_u_u.const_array);
  }
  failure_functor = new hash_const_cookie_functor(var, -1);
}

hash_const_struct::hash_const_struct(mu_state *m, mint_const *d,
				     int domain_size, cast_expr var)
{
  must = m;
  count = domain_size;
  // build the data_type domain...
  domain = new data_type[count];
  success_functors = new functor *[count];
  for (int i = 0; i < count; i++) {
    if (d[i]->kind != MINT_CONST_STRUCT)
      panic("Incorrect type in a domain of constants...\n");
    success_functors[i] = new hash_const_cookie_functor(var, i);
    domain[i] = build(d[i]->mint_const_u_u.const_struct);
  }
  failure_functor = new hash_const_cookie_functor(var, -1);
}

/*****************************************************************************/

hash_const::hash_const(mu_state *mu, mint_const *dom, int dom_size,
		       functor **success_funs, functor *failure_fun)
{
	mint_const_kind dom_mint_kind;
	int i;
	
	must = mu;
	domain = new data_type[dom_size];
	count = dom_size;
	success_functors = success_funs;
	failure_functor = failure_fun;
	
	if (count > 0) {
		dom_mint_kind = dom[0]->kind;
		for (i = 0; i < count; ++i)
			if (dom[i]->kind != dom_mint_kind)
				panic("Inconsistent constant types!");
		
		switch (dom[0]->kind) {
		case MINT_CONST_INT:
			for (i = 0; i < count; ++i)
				domain[i] = build(dom[i]->
						  mint_const_u_u.const_int);
			break;
		case MINT_CONST_FLOAT:
			for (i = 0; i < count; ++i)
				domain[i] = build(dom[i]->
						  mint_const_u_u.const_float);
			break;
		case MINT_CONST_CHAR:
			for (i = 0; i < count; ++i)
				domain[i] = build(dom[i]->
						  mint_const_u_u.const_char);
			break;
		case MINT_CONST_ARRAY:
			for (i = 0; i < count; ++i)
				domain[i] = build(dom[i]->
						  mint_const_u_u.const_array);
			break;
		case MINT_CONST_STRUCT:
			for (i = 0; i < count; ++i)
				domain[i] = build(dom[i]->
						  mint_const_u_u.const_struct);
			break;
		default:
			panic("Unrecognized MINT constant kind");
			break;
		}
	}
}

/*****************************************************************************/

void
mu_state::mu_discriminate(mint_const *domain,
			  mint_ref domain_itype,
			  int domain_size,
			  functor **success_functors,
			  functor *failure_functor)
{
	hash_const *hash_func;
	
	if (op & MUST_DECODE) {
		hash_func = new hash_const(this,
					   domain, domain_size,
					   success_functors, failure_functor);
		hash_func->hash();
		delete hash_func;
	}
	if (op & MUST_ENCODE) {
		assert(domain_size == 1);
		mu_encode_const(domain[0], domain_itype);
		success_functors[0]->func(this);
	}
	/* What about allocation-only or deallocation-only passes? */
}

