/*
 * Copyright (c) 1996, 1997 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 */

#include "suntcp-link.h"

#include <signal.h>

/* A signal handler is a pointer to function taking int returning void. */
typedef void (*signal_handler_t)(int);

/*****************************************************************************/

int flick_read_buf(FLICK_BUFFER *out_buf, int sock)
{
	/* Semantics of this call:
	   This call should block until it gets a message Reading a
	   message: on the first read, the first word received will be a
	   network order byte count, including the first word.  all
	   remaining reads should block until the entire message is
	   received.  */

	int real_buf_size = out_buf->real_buf_end - out_buf->real_buf_start;
	int bytes_read = 1; /* init to non-zero value */
#ifdef TRACK_FRAGMENTATION
	int mult = 0;
#endif
	if ((out_buf->buf_read == out_buf->real_buf_start)/* new buffer */
	    || (out_buf->buf_end >= out_buf->buf_read)) /* buf is empty */ {
		out_buf->buf_start = out_buf->buf_read = out_buf->real_buf_start;
		
		if ((bytes_read = read(sock, (char *)out_buf->buf_read,
				       (real_buf_size > READ_PACKET) ? READ_PACKET
				       : real_buf_size)) > 0)
			out_buf->buf_read += bytes_read;
	} else {
		/* go to the next message */
		out_buf->buf_start = out_buf->buf_end;
	}
	
	if (bytes_read > 0) {
		
		unsigned int buf_size = (ntohl(*(int *)out_buf->buf_start)
					 & 0x7fffffffL) + 4;
		out_buf->buf_end = out_buf->buf_start + buf_size;
		
		/* if only partial message received */
		if (out_buf->buf_end > out_buf->buf_read) {
			
			/* if the rest of the message won't fit */
			if (out_buf->buf_end > out_buf->real_buf_end) {
				
				/* copy the partial message into the
                                   beginning of the buffer */
				memcpy(out_buf->buf_start, out_buf->real_buf_start,
				       out_buf->buf_read - out_buf->buf_start);
				/* adjust buf_read */
				out_buf->buf_read -= out_buf->buf_start
						     - out_buf->real_buf_start;
				
				/* if the message is bigger than the buffer */
				if (buf_size > real_buf_size) {
					
					/* buf_start used as storage for
                                           out_buf->real_buf_start */
					out_buf->buf_start = realloc
							     (out_buf->
							      real_buf_start,
							      buf_size);
					
					/* adjust real_buf and buf values
                                           because of realloc */
					real_buf_size = buf_size;
					out_buf->real_buf_end =/* same as buf_end */
					out_buf->buf_end = out_buf->buf_start
							  + buf_size;
					out_buf->buf_read +=
						(out_buf->buf_start
						 - out_buf->real_buf_start);
					out_buf->real_buf_start =
						out_buf->buf_start;
				} else {
					out_buf->buf_start =
						out_buf->real_buf_start;
					out_buf->buf_end = out_buf->buf_start
							   + buf_size;
				}
			}
			do {
#ifdef TRACK_FRAGMENTATION
				mult++;
#endif
				if ((bytes_read = read(sock, (char *)out_buf
						       ->buf_read,
						       real_buf_size
						       - (out_buf->buf_read
							  - out_buf->real_buf_start)
					)) > 0)
					out_buf->buf_read += bytes_read;
			} /* until the whole message is read */
			while (bytes_read > 0 && out_buf->buf_end
			                         > out_buf->buf_read);
		}
		
	}
#ifdef TRACK_FRAGMENTATION
	if (mult)
		printf("Multiple (%d) reads on server!\n", mult);
#endif
	/* printbuf(out_buf, buf_offset); */
	if (bytes_read <= 0) {
		close(sock);
		return 0;
	}
	
	return 1;
}

void print_buf(int size, void *data)
{
	int i, j;
	for (j = 0; j < 2; j++) {
		for (i = 0; i < size; i++) {
			if (i % 16 == 0)
				printf("\n%06x: ", i);
			else if (i % 4 == 0)
				printf(" ");
			if (j)
				printf(" %c", (((char *)data)[i]) < ' ' ? ' ' : ((char *)data)[i]);
			else
				printf("%02x", ((char *)data)[i] & 0xff);
		}
	}
	printf("\n");
}

int flick_write_buf(FLICK_BUFFER *buf, int socket)
{
	signal_handler_t saved_sigpipe_handler;
	
	unsigned int size = (char *)buf->buf_current - (char *)buf->buf_start;
	int written = 0;
	int offset = 0;
	/* Set the size of the buffer */
	((int *)buf->buf_start)[0] = htonl((0x80000000 | size) - 4);
#if 0
	printf("Writing:");
	print_buf(size, buf->buf_start);
#endif
	
	/* Ignore SIGPIPE: writes on a socket with no reader. */
	saved_sigpipe_handler = signal(SIGPIPE, SIG_IGN);
	
	/* I'm not sure if this while loop is necessary... */
	while (size > 0) {
		int write_size = (size < WRITE_PACKET) ? size : WRITE_PACKET;
		written = write(socket,
				&((char *)buf->buf_start)[offset],
				write_size);
		
		if (written < 0)
			break;
		
		offset += written;
		size -= written;
	}

	/* Restore the old SIGPIPE handler. */
	(void) signal(SIGPIPE, saved_sigpipe_handler);
	
	if (size > 0) {
		perror("socket write failure");
		/* XXX --- We used to `close(socket)' here.  Gack! */
	}
	
	/* Return 1 (true) for success or 0 (false) for failure. */
	return (size == 0);
}

/* End of file. */

